import mne
import logging
import numpy as np
from datamatrix._datamatrix._seriescolumn import _SeriesColumn
from datamatrix import series as srs, NAN, operations as ops, functional as fnc
from functools import lru_cache
import autoreject
import os

logger = logging.getLogger('eeg_eyetracking_parser')


class PupilEpochs(mne.Epochs):
    """An Epochs class for the PupilSize channel. This allows baseline
    correction to be applied to pupil size, even though this channel is not a
    regular data channel.
    """
    
    def __init__(self, *args, **kwargs):
        mne.io.pick._PICK_TYPES_DATA_DICT['misc'] = True
        super().__init__(*args, **kwargs, picks='PupilSize')
        mne.io.pick._PICK_TYPES_DATA_DICT['misc'] = False
        
    def average(self, *args, **kwargs):
        mne.io.pick._PICK_TYPES_DATA_DICT['misc'] = True
        evoked = super().average(*args, **kwargs)
        mne.io.pick._PICK_TYPES_DATA_DICT['misc'] = False
        return evoked


@fnc.memoize(persistent=True)
def autoreject_epochs(*args, ar_kwargs=None, **kwargs):
    """A factory function that creates an Epochs() object, applies
    autorejection, and then returns it.
    
    __Important:__ This function uses persistent memoization, which means that
    the results for a given set of arguments are stored on disk and returned
    right away for subsequent calls. For more information, see
    <https://pydatamatrix.eu/memoization/>

    Parameters
    ----------
    *args: iterable
        Arguments passed to mne.Epochs()
    ar_kwargs: dict or None, optional
        Keywords to be passed to AutoReject(). If `n_interpolate` is not
        specified, a default value of [1, 4, 8, 16] is used.
    **kwargs: dict
        Keywords passed to mne.Epochs()

    Returns
    -------
    Epochs:
        An mne.Epochs() object with autorejection applied.
    """
    # autoreject is initially run on all EEG channels, and only afterwards are
    # channels picked (if any picks are specified)
    if 'picks' in kwargs:
        picks = kwargs['picks']
        del kwargs['picks']
    else:
        picks is None
    epochs = mne.Epochs(*args, preload=True, picks='eeg', **kwargs)
    if ar_kwargs is None:
        ar_kwargs = {}
    if 'n_interpolate' not in ar_kwargs:
        ar_kwargs['n_interpolate'] = [1, 4, 8, 16]
    ar = autoreject.AutoReject(**ar_kwargs)
    epochs = ar.fit_transform(epochs)
    if picks is not None:
        epochs.pick_channels(picks)
    return epochs


def epochs_to_series(dm, epochs, baseline_trim=None):
    """Takes an Epochs, PupilEpochs, or EpochsTFR object and converts it to a
    DataMatrix SeriesColumn. If a baseline has been specified in the epoch, it
    is applied to each row of the series separately. Rows where the mean
    baseline value (z-scored) is not within the range indicated by
    `baseline_trim` are set to `NAN`.

    Parameters
    ----------
    dm: DataMatrix
        A DataMatrix object to which the series belongs
    epochs: Epochs, PupilEpochs, or EpochsTFR
        The source object with the epoch data.
    baseline_trim: tuple of int, optional
        The range of acceptable baseline values. This refers to z-scores.

    Returns
    -------
    SeriesColumn
    """
    # Different objects have different ways to access the data
    if hasattr(epochs, 'get_data') and callable(epochs.get_data):
        a = epochs.get_data()
    elif hasattr(epochs, 'data'):
        a = epochs.data
    else:
        raise TypeError('invalid epochs object')
    # If the data is four dimensional, then it's probably a time-frequency
    # object where the second dimension reflects the frequency bands while
    # the third dimension reflects channels. We average over both.
    if len(a.shape) == 4:
        a = a.mean(axis=(1, 2))
    # If the data is three dimension, then we only average over channels.
    else:
        a = a.mean(axis=1)
    s = _SeriesColumn(dm, depth=a.shape[1])
    s._seq[epochs.metadata.index] = a
    if hasattr(epochs, 'baseline') and epochs.baseline is not None:
        start, end = epochs.baseline
        t = epochs._raw_times
        i = np.where((t >= start) & (t <= end))[0]
        bl_start = i[0]
        bl_end = i[-1]
        if baseline_trim is not None:
            bl = srs.reduce(s[:, bl_start : bl_end])
            for row_nr, (z, bl) in enumerate(zip(ops.z(bl), bl)):
                if z < baseline_trim[0] or z > baseline_trim[1]:
                    logging.debug(
                        f'setting trace to NAN because baseline out of bounds (bl={bl}, z={z})')
                    s[row_nr] = NAN
        s = srs.baseline(s, s, bl_start, bl_end)
    return s


def tfr_to_surface(dm, epochs):
    """Takes an EpochsTFR object and converts it to a DataMatrix
    MultiDimensionalColumn of two dimensions.

    Parameters
    ----------
    dm: DataMatrix
        A DataMatrix object to which the series belongs
    epochs: EpochsTFR
        The source object with the epoch data.

    Returns
    -------
    SeriesColumn
    """
    try:
        from datamatrix._datamatrix._multidimensionalcolumn import \
            _MultiDimensionalColumn
    except ImportError:
        raise Exception(
            'Your version of DataMatrix does not support multidimensional columns')
    # Different objects have different ways to access the data
    try:
        a = epochs.data
    except AttributeError:
        raise TypeError('invalid EpochsTFR object')
    if len(a.shape) != 4:
        raise TypeError('invalid epochs object')
    a = a.mean(axis=1)
    s = _MultiDimensionalColumn(dm, shape=a.shape[1:])
    s._seq[epochs.metadata.index] = a
    return s



import numpy as np
import mne
import os
from ._triggers import trial_trigger
from mne.preprocessing import ICA, annotate_muscle_zscore
import autoreject as ar
import json
import matplotlib.pyplot as plt


def rereference_channels(raw, ref_channels=['A1', 'A2']):
    """
    Re-reference EEG data to the mastoids

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    ref_channels: string, list of strings, optional
        The channels that you want to re-reference to

    Returns
    -------
    raw: instance of raw EEG data
        A raw re-referenced EEG data
    """
    raw.set_eeg_reference(ref_channels=ref_channels)
    raw.drop_channels(ref_channels)


def drop_unused_channels(raw, chan_name_pattern='Channel'):
    """
    Drop unused channels. The pattern is that they all start with "Channel"

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    chan_name_pattern: string, optional
        All the channels that start with chan_name pattern will be excluded as
        they contain no data

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with empty channels removed
    """
    empty_chan_names = [chan_name for chan_name in raw.info['ch_names']\
                        if chan_name.startswith(chan_name_pattern)]
    raw.drop_channels(empty_chan_names)


def create_eog_channels(raw,
                        eog_channels=['VEOGT', 'VEOGB', 'HEOGL', 'HEOGR']):
    """
    Create EOG channels by subtracting corresponding channels
    (VEOG = VEOGT - VEOGB)
    (HEOG = HEOGL - HEOGR)

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    eog_channels: list of strings, optional
        VEOG top, VEOG bottom, HEOG left and HEOG right

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with two EOG channels
    """
    chinds = [raw.ch_names.index(ichannel) for ichannel in eog_channels]
    veog = raw._data[chinds[0], :] - raw._data[chinds[1], :]
    heog = raw._data[chinds[2], :] - raw._data[chinds[3], :]
    veog = veog[np.newaxis, :]
    heog = heog[np.newaxis, :]
    info = mne.create_info(['VEOG', 'HEOG'], raw.info['sfreq'], ['eog', 'eog'])
    neweog = mne.io.RawArray(np.r_[veog, heog], info)
    raw.add_channels([neweog], force_update_info=True)
    raw.set_channel_types(dict(HEOG='eog', VEOG='eog'))
    raw.drop_channels(eog_channels)


def set_montage(raw, montage_name='standard_1020', plot=False):
    """
    Set Montage for EEG channel positions
    Here we use standard_1020 and thus electrode names and positions are
    according to international 10-20 system

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    montage_name: string, optional
        A montage that contains channel positions
    plot: bool, optional
        True if plotting is to be done. By default, no plotting

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with channel locations
    """
    montage = mne.channels.make_standard_montage(montage_name)
    if plot:
        montage.plot()
    if 'Digi' in raw.info['ch_names']:
        raw.set_channel_types({'Digi': 'stim'})
    raw.set_montage(montage, match_case=False)


def annotate_emg(raw, threshold=5, ch_type="eeg",
                 muscle_frequencies=[110, 140], plot=False,
                 preprocessing_path=None, subject_nr=None):
    """
    Annotate segments of data that represent muscle artifacts

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    threshold: float, optional
        The threshold in z-scores for marking data as muscle artifacts
    ch_type: string, optional
        Type of channels we select, here - eeg
    muscle_frequencies: list, optional
        Frequencies of bandpass filter, as muscle artifacts are noticable
        primarily at high frequencies
    plot: bool, optional
        True if plotting is to be done. By default, no plotting
    preprocessing_path: string, optional
        If the output is saved, specify the preprocessing path it will be saved
        to
    subject_nr: int, optional
        Specify subject_nr for saving to correct folder

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with segments of data annototed with regard to muscle
        artifacts
    """
    muscle_annotations, muscle_scores = annotate_muscle_zscore(
        raw, ch_type=ch_type, threshold=threshold,
        filter_freq=muscle_frequencies)
    raw.set_annotations(raw.annotations + muscle_annotations)
    if plot and preprocessing_path is not None:
        emg_dir = os.path.join(preprocessing_path, "EMG")
        subject_emg_dir = os.path.join(emg_dir, 'subject_' + str(subject_nr))
        if not os.path.exists(subject_emg_dir):
            os.makedirs(subject_emg_dir)
        fig, ax = plt.subplots()
        ax.plot(raw.times, muscle_scores)
        ax.axhline(y=threshold, color='blue')
        ax.set(xlabel='time in s', ylabel='zscore')
        plt.savefig(
            os.path.join(subject_emg_dir, 'muscle_annotations_zscore.png'))
        plt.show()


def notch_filter(raw, frequencies_remove=(50, 100, 150, 200)):
    """
    Removes power line noise generated by EEG system

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    frequencies_remove: float, array of float, optional
        Frequencies to remove from the data

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with certain frequencies that were filtered out
    """
    raw.notch_filter(frequencies_remove, picks=['eeg', 'eog'])


def band_pass_filter(raw, lf=0.1, hf=40, plot=False):
    """
    High-pass and low-pass filter for EEG data (band-pass filter)
    High-pass removes slow drifts
    Low-pass removes the short-term fluctuations

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    lf: float, optional
        Low-pass frequency edge
    hf: float, optional
        High-pass frequency edge
    plot: bool, optional
        True if plotting is to be done. By default, no plotting

    Returns
    -------
    raw: instance of raw EEG data
        A raw filtered EEG data
    """
    if plot:
        raw_unfiltered = raw.copy()
    raw.filter(l_freq=lf, h_freq=hf, picks=['eeg', 'eog'])
    if plot:
        for title, data in zip(['Unfiltered', 'Bandpass filtered'],
                               [raw_unfiltered, raw]):
            fig = data.plot_psd(fmax=min(250, raw.info['sfreq'] / 2))
            fig.suptitle(title)


def downsample_data(raw, events, srate=250):
    """
    Downsample EEG data for computational efficacy

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    events: tuple
        Event information as returned by `read_subject()`.
    srate: float, optional
        Sampling rate to convert to

    Returns
    -------
    raw: instance of raw EEG data
        A raw downsampled EEG data
    events: tuple
        Downsampled events.
    """
    events, event_id = events
    raw, events = raw.resample(srate, events=events, npad="auto")
    return raw, (events, event_id)


def autodetect_bad_channels(raw, events, plot=False, preprocessing_path=None, 
                            subject_nr=None, eeg_scaling=20e-5):
    """
    Detect bad channels using ransac algorithm

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    events: tuple
        Event information as returned by `read_subject()`.
    plot: bool, optional
        True if plotting is to be done. By default, no plotting
    preprocessing_path: string, optional
        If the output is saved, specify the preprocessing path it will be saved
        to
    subject_nr: int, optional
        Specify subject_nr for saving to correct folder
    eeg_scaling: dict, optional
        Scaling with which to visualize the data, in µV

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with bad channels detected (raw.info['bads])
    """
    raw.info['bads'] = []
    raw_bads = raw.copy()
    raw_bads.pick_types(eeg=True)
    trial_events = trial_trigger(events)
    median_duration = np.median(
        trial_events[1:, 0] - trial_events[:-1, 0]) * 1.5
    cue_epoch = mne.Epochs(raw_bads, trial_events, tmin=0, baseline=None,
                           tmax=median_duration / 1000, preload=True)
    ransac = ar.Ransac()
    ransac = ransac.fit(cue_epoch)
    raw.info['bads'] = ransac.bad_chs_
    if plot or preprocessing_path is not None:
        bad_dir = os.path.join(preprocessing_path, "Bad_channels")
        subject_bad_dir = os.path.join(
            bad_dir, 'subject_' + str(subject_nr))
        if not os.path.exists(subject_bad_dir):
            os.makedirs(subject_bad_dir)
        if plot:
            raw_bads.plot(butterfly=True, bad_color='r', scalings=eeg_scaling)
            plt.savefig(
                os.path.join(subject_bad_dir, 'raw_with_bads_ransac.png'))
            plt.show()
            plt.figure(figsize=(20, 5))
            plt.bar(range(len(raw_bads.ch_names)), ransac.bad_log.mean(0))
            plt.xticks(range(len(raw_bads.ch_names)), raw_bads.ch_names)
            plt.tight_layout()
            plt.savefig(os.path.join(subject_bad_dir, 'ransac_scores.png'))
            plt.show()
        if preprocessing_path is not None:
            bads_txt_file = os.path.join(subject_bad_dir, 'bads.txt')
            raw_bads_file = os.path.join(subject_bad_dir, 'raw_bads_eeg.fif')
            with open(bads_txt_file, 'w') as f:
                json.dump(raw.info['bads'], f)
            raw.save(raw_bads_file, overwrite=True)


def run_ica(raw, lf=1, sel_components='all', ica_method='picard', n_iter=500,
            random_state_set=97, preprocessing_path=None, subject_nr=None):
    """
    Run ICA (independent component analysis)
    Here we use a robust and fast 'picard' method by default as a fast and
    robust algorithm

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    lf: float, optional
        Low-pass frequency edge
    sel_components: string, int, float, optional
        Be default, 'all' meaning that number of components = number of
        channels
    ica_method: string, optional
        Choice of the following methods: ‘fastica’, ‘infomax’ or ‘picard’
    n_iter: int, optional
        Maximum number of iterations
    random_state_set:int, optional
        Seed for random generator
    preprocessing_path: string, optional
        If the output is saved, specify the preprocessing path it will be saved
        to
    subject_nr: int, optional
        Specify subject_nr for saving to correct folder

    Returns
    -------
    ica: instance of ICA
        A result of independent component analysis
    """
    raw_ica = raw.copy().filter(l_freq=lf, h_freq=None)

    if sel_components == 'all':
        picks = mne.pick_types(raw.info, eeg=True)
        n_sel_components = len(picks)
    else:
        n_sel_components = sel_components
    ica = ICA(n_components=n_sel_components, method=ica_method,
              max_iter=n_iter, random_state=random_state_set)
    ica.fit(raw_ica)
    if preprocessing_path is not None:
        ica_dir = os.path.join(preprocessing_path, "ICA")
        subject_ica_dir = os.path.join(
            ica_dir, 'subject_' + str(subject_nr))
        if not os.path.exists(subject_ica_dir):
            os.makedirs(subject_ica_dir)
        ica.save(os.path.join(subject_ica_dir, 'res_ica.fif'), overwrite=True)
    return ica


def auto_select_ica(raw, ica, plot=False, preprocessing_path=None,
                    subject_nr=None):
    """
    Select ICA components automatically by matching them to EOG channels

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    ica: instance of ICA
        A result of independent component analysis
    plot: bool, optional
        True if plotting is to be done. By default, no plotting
    preprocessing_path: string, optional
        If the output is saved, specify the preprocessing path it will be saved
        to
    subject_nr: int, optional
        Specify subject_nr for saving to correct folder

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with ICA components removed
    """
    eog_indices, eog_scores = ica.find_bads_eog(raw)
    ica.exclude = eog_indices
    ica.apply(raw)
    if preprocessing_path is None:
        return
    ica_dir = os.path.join(preprocessing_path, "ICA")
    subject_ica_dir = os.path.join(
        ica_dir, 'subject_' + str(subject_nr))
    if not os.path.exists(subject_ica_dir):
        os.makedirs(subject_ica_dir)
    if plot:
        for ieog in eog_indices:
            ica.plot_properties(raw, picks=ieog)
            plt.savefig(os.path.join(
                subject_ica_dir, str(ieog) + 'ica_component_properties.png'))
            plt.show()
        ica.plot_sources(raw, show_scrollbars=False)
        plt.savefig(os.path.join(subject_ica_dir, 'latent_sources_raw.png'))
        plt.show()
        ica.plot_scores(eog_scores)
        plt.savefig(os.path.join(subject_ica_dir, 'ica_scores.png'))
        plt.show()
    ica_txt_file = os.path.join(subject_ica_dir, 'ica_removed.txt')
    ica_components = ['ICA component ' + str(x) for x in eog_indices]
    with open(ica_txt_file, 'w') as f:
        json.dump(ica_components, f)


def interpolate_bads(raw, resetting=False):
    """
    Interpolate channels that we previously marked as bads

    Parameters
    ----------
    raw: instance of raw EEG data
        A raw EEG data
    resetting: bool, optional
        If False bads are still mentioned in 'info'

    Returns
    -------
    raw: instance of raw EEG data
        A raw EEG data with bad channels interpolated
    """
    raw.interpolate_bads(reset_bads=resetting)



import logging
from pathlib import Path
import numpy as np
import mne
from datamatrix import convert as cnv, io, functional as fnc
from datamatrix._datamatrix._seriescolumn import _SeriesColumn
from eyelinkparser import parse, defaulttraceprocessor
from ._triggers import trial_trigger, _parse_triggers, _validate_events
from . import _eeg_preprocessing as epp

logger = logging.getLogger('eeg_eyetracking_parser')


@fnc.memoize(persistent=True)
def read_subject(subject_nr, folder='data/', trigger_parser=None,
                 eeg_margin=30, min_sacc_dur=10, min_sacc_size=100,
                 min_blink_dur=10, blink_annotation='BLINK',
                 saccade_annotation='SACCADE', eeg_preprocessing=True,
                 save_preprocessing_output=True, plot_preprocessing=False,
                 eye_kwargs={}, downsample_data_kwargs={},
                 drop_unused_channels_kwargs={}, 
                 rereference_channels_kwargs={}, create_eog_channels_kwargs={},
                 set_montage_kwargs={}, annotate_emg_kwargs={},
                 band_pass_filter_kwargs={}, autodetect_bad_channels_kwargs={},
                 run_ica_kwargs={}, auto_select_ica_kwargs={},
                 interpolate_bads_kwargs={}):
    """Reads EEG, eye-tracking, and behavioral data for a single participant.
    This data should be organized according to the BIDS specification.
    
    EEG data is assumed to be in BrainVision data format (`.vhdr`, `.vmrk`,
    `.eeg`). Eye-tracking data is assumed to be in EyeLink data format (`.edf`
    or `.asc`). Behavioral data is assumed to be in `.csv` format.
    
    Metadata is taken from the behavioral `.csv` file if present, and from
    the eye-tracking data if not.
    
    __Important:__ This function uses persistent memoization, which means that
    the results for a given set of arguments are stored on disk and returned
    right away for subsequent calls. For more information, see
    <https://pydatamatrix.eu/memoization/>

    Parameters
    ----------
    subject_nr: int or sr
        The subject number to parse. If an int is passed, the subject number
        is assumed to be zero-padded to length two (e.g. '01'). If a string
        is passed, the string is used directly.
    folder: str, optional
        The folder in which the data is stored.
    trigger_parser: callable, optional
        A function that converts annotations to events. If no function is
        specified, triggers are assumed to be encoded by the OpenVibe
        acquisition software and to follow the convention for indicating
        trial numbers and event onsets as described in the readme.
    eeg_margin: int, optional
        The number of seconds after the last trigger to keep. The rest of the
        data will be cropped to save memory (in case long periods of extraneous
        data were recorded).
    min_sacc_dur: int, optional
        The minimum duration of a saccade before it is annotated as a
        BAD_SACCADE.
    min_sacc_size: int, optional
        The minimum size of a saccade (in pixels) before it is annotated as a
        saccade.
    min_blink_dur: int, optional
        The minimum duration of a blink before it is annotated as a blink.
    blink_annotation: str, optional
        The annotation label to be used for blinks. Use a BAD_ suffix to
        use blinks a bads annotations.
    saccade_annotation: str, optional
        The annotation label to be used for saccades. Use a BAD_ suffix to
        use saccades a bads annotations.
    eeg_preprocessing: bool or list, optional
        Indicates whether EEG preprocessing should be performed. If `True`,
        then all preprocessing steps are performed. If a list is passed, then
        only those steps are performed for which the corresponding function
        name is in the list (e.g. `['downsample_data', 'set_montage']`)
    save_preprocessing_output: bool, optional
        Indicates whether output generated during EEG preprocessing should be
        saved.
    plot_preprocessing: bool, optional
        Indicates whether plots should be shown during EEG preprocessing.
    eye_kwargs: dict, optional
        Optional keyword arguments to be passed onto the EyeLink parser. If
        traceprocessor is provided, a default traceprocessor is used with
        advanced blink reconstruction enabled and 10x downsampling.
    downsample_data_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    drop_unused_channels_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    rereference_channels_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    create_eog_channels_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    set_montage_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    annotate_emg_kwargs: dict, optional
         Passed as keyword arguments to corresponding preprocessing function.
    band_pass_filter_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    autodetect_bad_channels_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    run_ica_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    auto_select_ica_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    interpolate_bads_kwargs: dict, optional
        Passed as keyword arguments to corresponding preprocessing function.
    
    Returns
    -------
    tuple:
        A raw (EEG data), events (EEG triggers), metadata (a table with
        experimental variables) tuple.
    """
    if isinstance(subject_nr, int):
        subject_path = Path(folder) / Path('sub-{:02d}'.format(subject_nr))
    else:
        subject_path = Path(folder) / Path('sub-{}'.format(subject_nr))
    logger.info(f'reading subject data from {subject_path}')
    raw, events = _read_eeg_data(subject_path / Path('eeg'), trigger_parser,
                                 eeg_margin)
    metadata = _read_beh_data(subject_path / Path('beh'))
    eye_path = subject_path / Path('eyetracking')
    dm = _read_eye_data(eye_path, metadata, eye_kwargs)
    if dm is not None and raw is not None:
        dm = _merge_eye_and_eeg_data(eye_path, raw, events, dm, min_sacc_dur,
                                     min_sacc_size, min_blink_dur,
                                     blink_annotation, saccade_annotation,
                                     eye_kwargs)
    if metadata is None and dm is not None:
        metadata = _dm_to_metadata(dm)
    if eeg_preprocessing:
        # Preprocessing output can be saved for visual inspection. By default
        # this is done in a subfolder of the participant data folder. This
        # only applies to certain preprocessing steps, i.e. those functions
        # that take a `preprocessing_path` keyword.
        if save_preprocessing_output:
            preprocessing_path = subject_path / Path('preprocessing')
            logger.info(f'saving preprocessing output to {preprocessing_path}')
            autodetect_bad_channels_kwargs['preprocessing_path'] = \
                preprocessing_path
            run_ica_kwargs['preprocessing_path'] = preprocessing_path
            auto_select_ica_kwargs['preprocessing_path'] = preprocessing_path
            annotate_emg_kwargs['preprocessing_path'] = preprocessing_path
        if plot_preprocessing:
            logger.info(f'creating plots during preprocessing')
            set_montage_kwargs['plot'] = True
            band_pass_filter_kwargs['plot'] = True
            autodetect_bad_channels_kwargs['plot'] = True
            auto_select_ica_kwargs['plot'] = True
            annotate_emg_kwargs['plot'] = True
        if isinstance(eeg_preprocessing, bool):
            eeg_preprocessing = [
                'drop_unused_channels',
                'rereference_channels',
                'annotate_emg',
                'downsample_data',
                'create_eog_channels',
                'set_montage',
                'band_pass_filter',
                'autodetect_bad_channels',
                'ica',
                'interpolate_bads']
        if 'drop_unused_channels' in eeg_preprocessing:
            logger.info('dropping unused channels')
            epp.drop_unused_channels(raw, **drop_unused_channels_kwargs)
        else:
            logger.info('*not* dropping unused channels')
        if 'rereference_channels' in eeg_preprocessing:
            logger.info('re-referencing channels')
            epp.rereference_channels(raw, **rereference_channels_kwargs)
        else:
            logger.info('*not* re-referencing channels')
        
        if 'annotate_emg' in eeg_preprocessing:
            logger.info('applying muscle artifact detection')
            epp.annotate_emg(raw, **annotate_emg_kwargs)
        else:
            logger.info('*not* applying muscle artifact detection')
        # Operations that require a high sampling rate, notable EMG annotation
        # should be done before downsampling. Everything else should be done
        # afterwards to save memory and time.
        if 'downsample_data' in eeg_preprocessing:
            logger.info('downsampling')
            raw, events = epp.downsample_data(raw, events,
                                              **downsample_data_kwargs)
        else:
            logger.info('*not* downsampling')
        if 'create_eog_channels' in eeg_preprocessing:
            logger.info('creating EOG channels')
            epp.create_eog_channels(raw, **create_eog_channels_kwargs)
        else:
            logger.info('*not* creating EOG channels')
        if 'set_montage' in eeg_preprocessing:
            logger.info('setting montage')
            epp.set_montage(raw, **set_montage_kwargs)
        else:
            logger.info('*not* setting montage')
        if 'band_pass_filter' in eeg_preprocessing:
            logger.info('applying band-pass filter')
            epp.band_pass_filter(raw, **band_pass_filter_kwargs)
        else:
            logger.info('*not* applying band-pass filter')
        if 'autodetect_bad_channels' in eeg_preprocessing:
            logger.info('autodetecting bad channels')
            epp.autodetect_bad_channels(raw, events,
                                        **autodetect_bad_channels_kwargs)
        else:
            logger.info('*not* autodetecting bad channels')
        if 'ica' in eeg_preprocessing:
            logger.info('running ICA')
            ica = epp.run_ica(raw, **run_ica_kwargs)
            logger.info('autoselect ICA')
            epp.auto_select_ica(raw, ica, **auto_select_ica_kwargs)
        else:
            logger.info('*not* running ICA')
        if 'interpolate_bads' in eeg_preprocessing:
            logger.info('interpolating bad channels')
            epp.interpolate_bads(raw, **interpolate_bads_kwargs)
        else:
            logger.info('*not* interpolating bad channels')
    return raw, events, metadata


def _dm_to_metadata(dm):
    """Convert dm to pandas dataframe but remove seriescolumns to save memory.
    """
    return cnv.to_pandas(dm[
        [col for name, col in dm.columns if not isinstance(col, _SeriesColumn)]
    ])


def _merge_eye_and_eeg_data(eye_path, raw, events,
                            dm, min_sacc_dur, min_sacc_size, min_blink_dur,
                            blink_annotation, saccade_annotation, eye_kwargs):
    """Add the eye data as channels to the EEG data. In addition, add blinks
    and saccades as annotations.
    """
    logger.info(f'merging eye-tracking and eeg data')
    # First read the eye-tracking data again, this time without downsampling
    # and without splitting the data into separate epochs. We do this, so that
    # we can merge this big dataset as channels into the EEG data.
    def only_trial(name): return name == 'trial'
    if 'traceprocessor' not in eye_kwargs:
        logger.info('no traceprocessor specified, using default')
        eye_kwargs['traceprocessor'] = defaulttraceprocessor(
            blinkreconstruct=True, mode='advanced')
    bigdm = parse(folder=eye_path, trialphase='trial', phasefilter=only_trial,
                  **eye_kwargs)
    # Valid rows are those rows in which there is an epoch 1. Since the epochs
    # are only parsed in dm (not in bigdm), we use indices in dm to filter both
    # dm and bigdm. The result is that dm and bigdm only contain rows that
    # should correspond to trials in the EEG recording.
    if '' in dm.t_onset_1:
        valid_rows = dm[dm[dm.t_onset_1] != '']
        logger.warning(
            f'ignoring {len(dm) - len(valid_rows)} rows of eye-tracking data without epoch 1')
        bigdm = bigdm[valid_rows]
        dm = dm[valid_rows]
    # Check for missing EEG triggers. If the first trigger is missing, a
    # ValueError is raised. If any other trigger is missing, it is marked and
    # removed also from dm and bigdm.
    triggers = trial_trigger(events)[:, 2]
    if triggers[0] != 128:
        raise ValueError(
            f'The first trial trigger is {triggers[0]}, should be 128')
    rows = list(range(len(dm)))
    for i, (tr1, tr2) in enumerate(zip(triggers[:-1], triggers[1:])):
        if tr2 - tr1 not in (1, -127):
            rows.remove(i + 1)
            logger.warning(f'missing trial trigger for trial {i + 1}')
    dm = dm[rows]
    bigdm = bigdm[rows]
    # Check if final triggers are missing from the EEG. If so, these are
    # truncated also from dm and bigdm. We also exclude the last trial,
    # because this last trial may not have been completed. (We don't check
    # this, but simply err on the side of caution.)
    missing = len(dm) - len(triggers)
    if missing > 0:
        logger.warning(
            f'final {missing} triggers missing from recording, truncating eye data with one extra trial because the last trial may be incomplete too')
        dm = dm[:-missing - 1]
        bigdm = bigdm[:-missing - 1]
        last_trigger_index = np.where(events[0][:, 2] >= 128)[0][-1]
        events = events[0][:last_trigger_index - 1], events[1]
    # Now double-check that EEG and eye-tracking data match in length
    n_trials_eeg = sum(events[0][:, 2] >= 128)
    assert(n_trials_eeg == len(dm))
    assert(n_trials_eeg == len(bigdm))
    logger.info(f'eeg data and metata have matching length')
    # The start of the trial as recorded by the EEG start-trial marker and the
    # eyelink start_trial message are usually a little offset. However, the
    # EEG epoch triggers should *not* be offset relative to the eye-tracking
    # start_phase messages. (If they are, the experiment is not correctly
    # programmed). Therefore, we use the first epoch trigger to determine the
    # offset of the eye-movement data relative to the EEG data. We add this to
    # both the eye-tracking dms as the `eye_offset` column.
    for i, (row, bigrow) in enumerate(zip(dm, bigdm)):
        eye_t0 = bigrow.ttrace_trial[0]
        eye_t1 = row.t_onset_1 - eye_t0
        triggers = events[0]
        trigger_index = np.where(triggers[:, 2] >= 128)[0][i]
        eeg_t0, eeg_t1 = triggers[trigger_index:trigger_index + 2][:, 0]
        eeg_t1 -= eeg_t0
        eye_offset = eeg_t1 - eye_t1
        bigrow.eye_offset = eye_offset
        dm.eye_offset = eye_offset
    logger.info(f'trial offset (eye - eeg): {dm.eye_offset.mean}')
    # Now add gaze coordinates and pupil size as channels to the raw object.
    # Periods of missing data (in between recordings) is set to the median.
    logger.info('adding GazeX, GazeY, and PupilSize channels')
    data = np.empty((3, len(raw)), dtype=float)
    data[:] = np.nan
    trialdepth = bigdm.ptrace_trial.depth
    for (timestamp, _, code), row in zip(trial_trigger(events), bigdm):
        timestamp += row.eye_offset
        if len(raw) <= timestamp + trialdepth:
            logger.warning('eeg recording ended before last eye tracking trial')
            continue
        data[0, timestamp: timestamp + trialdepth] = row.xtrace_trial
        data[1, timestamp: timestamp + trialdepth] = row.ytrace_trial
        data[2, timestamp: timestamp + trialdepth] = row.ptrace_trial
    data[0, np.isnan(data[0])] = np.nanmedian(data[0])
    data[1, np.isnan(data[1])] = np.nanmedian(data[1])
    data[2, np.isnan(data[2])] = np.nanmedian(data[2])
    ch_names = ['GazeX', 'GazeY', 'PupilSize']
    ch_types = 'misc'
    info = mne.create_info(
        ch_names=ch_names,
        sfreq=raw.info['sfreq'],
        ch_types=ch_types)
    tmp = mne.io.RawArray(data, info)
    raw.add_channels([tmp])
    # Finally, get the blinks and saccades from the eye-tracking data and add
    # them as BAD annotations to the raw.
    logger.info('adding blink and saccade annotations')
    onset = []
    duration = []
    description = []
    for (timestamp, _, code), row in zip(trial_trigger(events), bigdm):
        timestamp += row.eye_offset
        t0 = row.ttrace_trial[0]
        for st, et in zip(row.blinkstlist_trial, row.blinketlist_trial):
            if np.isnan(st):
                break
            dur = et - st
            if dur < min_blink_dur:
                continue
            start = st - t0
            onset.append((timestamp + start) / 1000)
            duration.append(dur / 1000)
            description.append(blink_annotation)
        # Saccades are deduced through the end of fixation n and the start of
        # fixation n + 1
        for st, sx, sy, et, ex, ey in zip(
                row.fixetlist_trial[:-1],
                row.fixxlist_trial[:-1],
                row.fixylist_trial[:-1],
                row.fixstlist_trial[1:],
                row.fixxlist_trial[1:],
                row.fixylist_trial[1:]):
            if np.isnan(et):
                break
            dur = et - st
            if dur < min_sacc_dur:
                continue
            start = st - t0
            size = ((sx - ex) ** 2 + (sy - ey) ** 2) ** .5
            if size < min_sacc_size:
                continue
            onset.append((timestamp + start) / 1000)
            duration.append(dur / 1000)
            description.append(saccade_annotation)
    annotations = mne.Annotations(
        onset=onset,
        duration=duration,
        description=description)
    raw.set_annotations(raw.annotations + annotations)
    return dm


def _read_eeg_data(eeg_path, trigger_parser, margin):
    """Reads eeg data and returns a raw, events tuple. If no eeg data is found,
    None, None is returned.
    """
    if not eeg_path.exists():
        logger.info('no eeg data detected')
        return None, None
    vhdr_path = list(eeg_path.glob('*.vhdr'))[0]
    logger.info(f'loading eeg data from {vhdr_path}')
    raw = mne.io.read_raw_brainvision(vhdr_path, preload=True)
    if raw.info['sfreq'] != 1000:
        logger.warning(
            f'sampling rate is {raw.info["sfreq"]} Hz, resampling to 1000 Hz.')
        raw = raw.resample(1000)
    logger.info('creating events from annotations')
    events = mne.events_from_annotations(raw,
        _parse_triggers if trigger_parser is None else trigger_parser)
    if margin is not None:
        end = min(raw.times[-1], events[0][:, 0][-1] / 1000 + margin)
        logger.info(f'trimming eeg to 0 - {end} s')
        raw.crop(0, end)
    logger.info('validating events')
    _validate_events(events)
    logger.info('creating annotations from events')
    raw.set_annotations(
        mne.annotations_from_events(
            events[0],
            sfreq=raw.info['sfreq']))
    return raw, events


def _read_beh_data(beh_path):
    """Reads behavioral data and returns it as a DataFrame. None is returned
    if no data is found.
    """
    if not beh_path.exists():
        logger.info('no behavioral data detected')
        return None
    csv_path = list(beh_path.glob('*.csv'))[0]
    logger.info('loading behavioral data from {csv_path}')
    return cnv.to_pandas(io.readtxt(csv_path))


def _read_eye_data(eye_path, metadata, kwargs):
    """Reads eye-tracking data and returns it as a DataMatrix. None is returned
    if no data is found.
    """
    if not eye_path.exists():
        logger.info('no eye data detected')
        return None
    logger.info('loading eye data from {eye_path}')
    return parse(folder=eye_path, maxtracelen=1, pupil_size=False,
                 gaze_pos=False)



import numpy as np
from mne.time_frequency import tfr_morlet, tfr_multitaper


def run_morlet(epochs, frequency_start, frequency_end, frequency_step=1, pick_channels=None):
    """
    Run time-frequency decomposition using Morlet wavelets

    Parameters
    ----------
    epochs: Epochs
        The source object with the epoch data.
    frequency_start: int
        The beginning of frequencies you are interested in (in Hz)
    frequency_end: int
        The end of frequencies you are interested in (in Hz)
    frequency_step: int, optional
        Step with which frequencies will be analyzed. By default it is set to 1,
        meaning that each frequency from beginning to end will be taken (in Hz)
    pick_channels: str, list of strings, optional
        The channels you want to compute time-frequency for. By default, it is set to None,
        meaning that all channels will be used

    Returns
    -------
    average_power: average TFR
        Average power from the specified frequencies and channels
    """
    frequencies = np.arange(frequency_start, frequency_end, frequency_step)
    n_cycles = frequencies / 2.
    if pick_channels:
        sel_channels = [epochs.ch_names.index(pick_channels)]
    else:
        sel_channels = None

    average_power = tfr_morlet(epochs, freqs=frequencies, return_itc=False, n_cycles=n_cycles, picks=sel_channels)
    return average_power

def run_multitaper(epochs, frequency_start, frequency_end, time_bandwidth, frequency_step=1):
    """
    Run time-frequency decomposition using DPSS tapers

    Parameters
    ----------
    epochs: Epochs
        The source object with the epoch data.
    frequency_start: int
        The beginning of frequencies you are interested in (in Hz)
    frequency_end: int
        The end of frequencies you are interested in (in Hz)
    time_bandwidth: float
        Number of tapers (time_bandwidth-1). If None, will be set to 4 (3 tapers).
    frequency_step: int, optional
        Step with which frequencies will be analyzed. By default it is set to 1,
        meaning that each frequency from beginning to end will be taken (in Hz)

    Returns
    -------
    average_power: average TFR
        Average power from the specified frequencies and channels
    """
    frequencies = np.arange(frequency_start, frequency_end, frequency_step)
    n_cycles = frequencies / 2.

    average_power = tfr_multitaper(epochs, freqs=frequencies, return_itc=False,
                                   time_bandwidth=time_bandwidth, n_cycles=n_cycles)
    return average_power



"""A set of functions to facilitate working with braindecode"""
import random
import mne
import copy
import itertools
import logging
from collections.abc import Sequence
import collections
collections.Sequence = Sequence  # for compatibility with Skorch
from matplotlib import pyplot as plt
import numpy as np
from scipy.stats import mode
import pandas as pd
from matplotlib.lines import Line2D
from sklearn.metrics import confusion_matrix
import torch
from datamatrix import functional as fnc, DataMatrix, operations as ops, \
    SeriesColumn, convert as cnv
from skorch.callbacks import LRScheduler
from skorch.helper import predefined_split
from braindecode.datasets import create_from_mne_epochs
from braindecode.visualization import plot_confusion_matrix
from braindecode import EEGClassifier
from braindecode.training import CroppedLoss
from braindecode.util import set_random_seeds
from braindecode.models import ShallowFBCSPNet, to_dense_prediction_model, \
    get_output_shape
from braindecode.preprocessing import Preprocessor, \
    exponential_moving_standardize
from . import read_subject, epoch_trigger

logger = logging.getLogger('eeg_eyetracking_parser')


@fnc.memoize(persistent=True)
def decode_subject(read_subject_kwargs, factors, epochs_kwargs, trigger,
                   epochs_query='practice == "no"', epochs=4, window_size=200,
                   window_stride=1, n_fold=4, crossdecode_factors=None,
                   patch_data_func=None, read_subject_func=None):
    """The main entry point for decoding a subject's data.
    
    Parameters
    ----------
    read_subject_kwargs: dict
        A dict with keyword arguments that are passed to eet.read_subject() to
        load the data. Additional preprocessing as specified in
        `preprocess_raw()` is applied afterwards.
    factors: str or list of str
        A factor or list of factors that should be decoded. Factors should be
        str and match column names in the metadata.
    epochs_kwargs: dict, optional
        A dict with keyword arguments that are passed to mne.Epochs() to
        extract the to-be-decoded epoch.
    trigger: int
        The trigger code that defines the to-be-decoded epoch.
    epochs_query: str, optional
        A pandas-style query to select trials from the to-be-decoded epoch. The
        default assumes that there is a `practice` column from which we only
        want to have the 'no' values, i.e. that we want exclude practice
        trials.
    epochs: int, optional
        The number of training epochs, i.e. the number of times that the data
        is fed into the model. This should be at least 2.
    window_size_samples: int, optional
        The length of the window to sample from the Epochs object. This should
        be slightly shorter than the actual Epochs to allow for jittered
        samples to be taken from the purpose of 'cropped decoding'.
    window_stride_samples: int, optional
        The number of samples to jitter around the window for the purpose of
        cropped decoding.
    n_fold: int, optional
        The total number of splits (or folds). This should be at least 2.
    crossdecode_factors: str or list of str, optional
        A factor or list of factors that should be decoded during tester. If
        provided, the classifier is trained using the factors specified in
        `factors` and tested using the factors specified in
        `crossdecode_factors`. In other words, specifying this keyword allow
        for crossdecoding.
    patch_data_func: callable or None, optional
        If provided, this should be a function that accepts a tuple of
        `(raw, events, metadata)` as returned by `read_subject()` and also
        returns a tuple of `(raw, events, metadata)`. This function can modify
        aspects of the data before decoding is applied.
    read_subject_func: callable or None, optional
        If provided, this should be a function that accepts keywords as
        provided through the `read_subject_kwargs` argument, and returns a
        tuple of `(raw, events, metadata)`. If not provided, the default
        `read_subject()` function is used.

    Returns
    -------
    DataMatrix
        Contains the original metadata plus four additional columns:
        
        - `braindecode_label` is a numeric label that corresponds to the
          to-be-decoded factor, i.e. the ground truth
        - `braindecode_prediction` is the predicted label
        - `braindecode_correct` is 1 for correct predictions and 0 otherwise
        - `braindecode_probabilities` is a SeriesColumn with the predicted
          probabilities for each label. The prediction itself corresponds to
          the index with the highest probability.
    """
    if not isinstance(epochs, int) or epochs < 2:
        raise ValueError('epochs should >= 2')
    if not isinstance(n_fold, int) or epochs < 2:
        raise ValueError('n_fold should >= 2')
    dataset, labels, metadata = read_decode_dataset(
        read_subject_kwargs, factors, epochs_kwargs, trigger, epochs_query,
        window_size=window_size, window_stride=window_stride,
        patch_data_func=patch_data_func, read_subject_func=read_subject_func)
    if crossdecode_factors is not None:
        cd_dataset, labels, metadata = read_decode_dataset(
            read_subject_kwargs, crossdecode_factors, epochs_kwargs, trigger,
            epochs_query, window_size=window_size, window_stride=window_stride,
            patch_data_func=patch_data_func,
            read_subject_func=read_subject_func)
    n_conditions = len(labels)
    logger.info(f'decoding {n_conditions} labels')
    predictions = DataMatrix(length=0)
    for fold in range(n_fold):
        train_data, test_data = _split_dataset(dataset, fold=fold,
                                               n_fold=n_fold)
        if crossdecode_factors is not None:
            _, test_data = _split_dataset(cd_dataset, fold=fold, n_fold=n_fold)
        n_train_conditions = len(set(d.y[0] for d in train_data.datasets))
        if n_train_conditions != n_conditions:
            raise ValueError('Some labels are missing from training set')
        n_test_conditions = len(set(d.y[0] for d in test_data.datasets))
        if n_test_conditions != n_conditions:
            raise ValueError('Some labels are missing from testing set')
        clf = train(train_data, test_data, epochs=epochs)
        # We can unbalance the data after training to save time and to make the
        # cell counts match again
        _unbalance_dataset(test_data)
        # We want to know which trial was predicted to have which label. For
        # that reason, we create a datamatrix with true and predicted labels.
        # These are not in the original order, so we also store timestamps
        # so that later we can sort the datamatrix back into the original order
        y_prob = clf.predict_proba(test_data)  # probabilities for each label
        y_pred = np.argmax(y_prob, axis=1)     # the winner
        resized_pred = y_pred.copy()
        resized_pred.resize(
            (len(test_data.datasets), len(test_data.datasets[0])))
        y_prob.resize((len(test_data.datasets), len(test_data.datasets[0]),
                       n_conditions))
        fold_predictions = DataMatrix(length=len(test_data.datasets))
        fold_predictions.y_true = [d.y[0] for d in test_data.datasets]
        fold_predictions.y_pred = mode(
            resized_pred, axis=1, keepdims=True)[0].flatten()
        fold_predictions.y_prob = SeriesColumn(depth=y_prob.shape[-1])
        fold_predictions.y_prob = y_prob.mean(axis=1)
        fold_predictions.timestamp = [
            d.windows.metadata.i_start_in_trial[0]
            for d in test_data.datasets
        ]
        predictions <<= fold_predictions
    # Add the true and predicted labels as new columns to the metadata
    predictions = ops.sort(predictions, by=predictions.timestamp)
    dm = cnv.from_pandas(metadata)
    dm.braindecode_label = predictions.y_true
    dm.braindecode_prediction = predictions.y_pred
    dm.braindecode_probabilities = predictions.y_prob
    dm.braindecode_correct = 0
    dm.braindecode_correct[
        dm.braindecode_label == dm.braindecode_prediction] = 1
    return dm


def read_decode_dataset(read_subject_kwargs, factors, epochs_kwargs, trigger,
                        epochs_query='practice == "no"', lesion=None,
                        window_size=200, window_stride=1,
                        patch_data_func=None, read_subject_func=None):
    """Reads a dataset and converts it to a format that is suitable for
    braindecode.
    """
    raw, events, metadata = (
        read_subject(**read_subject_kwargs)
        if read_subject_func is None else
        read_subject_func(**read_subject_kwargs))
    if patch_data_func is not None:
        raw, events, metadata = patch_data_func(raw, events, metadata)
    _preprocess_raw(raw)
    epochs = mne.Epochs(raw, epoch_trigger(events, trigger),
                        metadata=metadata, **epochs_kwargs)
    n_dropped = len(metadata) - len(epochs)
    if n_dropped:
        logger.info(f'{n_dropped} epochs were dropped')
        metadata = metadata.iloc[epochs.selection]
    epochs = epochs[epochs_query]
    metadata = metadata.query(epochs_query)
    if isinstance(lesion, tuple):
        epochs._data[:, :, lesion[0]:lesion[1]] = 0
    elif isinstance(lesion, str):
        epochs._data[:, epochs.ch_names.index(lesion)] = 0
    dataset, labels = _build_dataset(
        epochs, metadata, factors, window_size_samples=window_size,
        window_stride_samples=window_stride)
    return dataset, labels, metadata


def train(train_set, test_set=None, epochs=4, batch_size=32, lr=0.000625,
          weight_decay=0, predict_nonlinearity='auto'):
    """Trains a classifier based on a training set. If a test set is provided,
    validation metrics are shown during training. However, for proper testing,
    the testing data should be predicted separately after training.

    Parameters
    ----------
    train_set: BaseDataSet
        The training data
    test_set: BaseDataSet, optional
        The testing (validation) data
    epochs: int, optional
        The number of training epochs, i.e. the number of times that the data
        is fed into the model
    batch_size: int, optional
        The number of observations that are fed into the model simultaneously
    lr: float, optional
        The learning rate. The default value is taken from the braindecode
        tutorials for the shallow network
    weight_decay: float, optional
        The weight decay. The default value is taken from the braindecode
        tutorials for the shallow network
    predict_nonlinearity: str, None, or callable
        See EEGClassifier docs
        
    Returns
    -------
    EEGClassifier
        A trained EEG classifier
    """
    # The number of unique y (code) values
    n_classes = len(set([d.y[0] for d in train_set.datasets]))
    # Number of electrode channels is hidden deep inside the data
    n_chans = train_set.datasets[0].windows.get_data().shape[1]
    input_window_samples = train_set.datasets[0].windows.get_data().shape[2]
    model = ShallowFBCSPNet(
        n_chans,
        n_classes,
        input_window_samples=input_window_samples,
        final_conv_length='auto',
    )
    if torch.cuda.is_available():
        logger.info('enabling cuda for gpu acceleration')
        model.cuda()
    else:
        logger.info('cuda is not available, not enabling gpu acceleration')
    to_dense_prediction_model(model)
    if test_set is None:
        train_split = None
        callbacks = None
    else:
        train_split = predefined_split(test_set)
        callbacks = [
            "accuracy", ("lr_scheduler",
                         LRScheduler('CosineAnnealingLR', T_max=epochs - 1)),
        ]
    clf = EEGClassifier(
        model,
        cropped=True,
        criterion=CroppedLoss,
        criterion__loss_function=torch.nn.functional.nll_loss,
        optimizer=torch.optim.AdamW,
        train_split=train_split,
        optimizer__lr=lr,
        optimizer__weight_decay=weight_decay,
        iterator_train__shuffle=True,
        batch_size=batch_size,
        callbacks=callbacks,
        predict_nonlinearity=predict_nonlinearity
    )
    clf.fit(train_set, y=None, epochs=epochs)
    return clf


def summarize_confusion_matrix(factors, confusion_mat):
    """Determines overall classification accuracy, as well as classification
    accuracy per factor for a confusion matrix. If there is more than one
    factor, each factor is assumed to have two levels.

    Parameters
    ----------
    factors: list
        A list of factor labels
    confusion_mat: array
        A confusion matrix as returned by build_confusion_matrix()

    Returns
    -------
    list
        A list of accuracies, where the first element is the overall accuracy
        and subsequent elements correspond to the factors.
    """
    if len(factors) == 1:
        model = np.identity(confusion_mat.shape[0])
        return [100 * np.sum(model * confusion_mat) / np.sum(confusion_mat)]
    model = np.identity(2 ** len(factors))
    accuracies = []
    acc = 100 * np.sum(model * confusion_mat) / np.sum(confusion_mat)
    accuracies.append(acc)
    for i, factor in enumerate(factors):
        r1 = 2 ** (len(factors) - i - 1)
        model = np.identity(2).repeat(r1, axis=0).repeat(r1, axis=1)
        r2 = 2 ** i
        model = np.block([[model] * r2] * r2)
        acc = 100 * np.sum(model * confusion_mat) / np.sum(confusion_mat)
        accuracies.append(acc)
    return accuracies


def build_confusion_matrix(labels, predictions):
    """Creates a confusion matrix based on the results of `decode_subject()`.
    
    Parameters
    ----------
    labels: DataMatrix column
        A column with true labels (`dm.braindecode_label`)
    predictions: DataMatrix column
        A column with predicted labels or probabilities for all labels
        (`dm.braindecode_prediction` or `dm.braindecode_probabilities`). 
    
    Returns
    -------
    array:
        A confusion matrix with the true labels as the first axis and the
        predicted labels (or summed probabilities of the predicted labels) as
        the second axis.
    """
    if hasattr(predictions, 'depth'):
        cm = np.zeros((labels.count, labels.count), dtype=float)
        for label, prediction in zip(labels, predictions):
            cm[label] += prediction
    else:
        cm = np.zeros((labels.count, labels.count), dtype=int)
        for label, prediction, ldm in ops.split(labels, predictions):
            cm[label, prediction] = len(ldm)
    return cm


def _preprocess_raw(raw, l_freq=4, h_freq=30, factor_new=1e-3,
                   init_block_size=1000):
    """Preprocesses the raw object such that is useful for decoding. The main
    criteria seem to be that high and low frequencies are removed, and that
    the signal is normalized (0 mean and 1 std). Based on the preprocessing
    steps in the braindecode tutorials.
    """
    raw.filter(l_freq=l_freq, h_freq=h_freq)
    Preprocessor(exponential_moving_standardize, factor_new=factor_new,
                 init_block_size=init_block_size).apply(raw)
                 

def _split_epochs(epochs, metadata, factors):
    """Splits an Epochs object based on several factors, which should
    correspond to columns in the metadata.

    Parameters
    ----------
    epochs: Epochs
        The Epochs object to split
    metadata: DataFrame
        The metadata that belongs to epochs
    factors: str or list
        A factor or list of factors by which the data should be split.

    Returns
    -------
    tuple
        A tuple in which the first element is a list of Epochs objects, and the
        second element a list of str labels that define the Epochs objects.
    """
    if isinstance(factors, str):
        factors = [factors]
    subsets = []
    labels = []
    for code, values in enumerate(
            itertools.product(*[np.unique(metadata[f]) for f in factors])):
        subset = epochs
        label = []
        for value, factor in zip(values, factors):
            label.append(f'{factor}={value}')
            if isinstance(value, str):
                value = f"'{value}'"
            subset = subset[f"{factor} == {value}"]
        subset.events[:, 2] = code
        labels.append('\n'.join(label))
        subsets.append(subset)
    return subsets, labels


def _split_dataset(dataset, n_fold=4, fold=0):
    """Split a dataset into training and testing sets using interleaved
    splitting.
    
    Parameters
    ----------
    n_fold: int
        The total number of splits (or folds)
    fold: int
        The current split number
        
    Returns
    -------
    tuple
        A (training set, test set) tuple
    """
    splitted = dataset.split({
        'train': [i for i in range(len(dataset.datasets)) if i % n_fold != fold],
        'test': [i for i in range(len(dataset.datasets)) if i % n_fold == fold]
    })
    return splitted['train'], splitted['test']


def _balance_dataset(dataset):
    """Makes sure that a dataset contains an equal number of observations for
    each label by randomly duplicating observations from labels that have too
    few observations. This modifies the dataset in-place.
    
    Parameters
    ----------
    dataset
    """
    n_codes = {}
    for d in dataset.datasets:
        code = d.y[0]
        if code not in n_codes:
            n_codes[code] = 0
        n_codes[code] += 1
    max_count = max(n_codes.values())
    for code, count in n_codes.items():
        n_add = max_count - count
        code_datasets = [d for d in dataset.datasets if d.y[0] == code]
        for i in range(n_add):
            copied_dataset = copy.deepcopy(random.choice(code_datasets))
            copied_dataset._is_balance_copy = True
            dataset.datasets.append(copied_dataset)
        if n_add:
            logger.info(
                f'adding {n_add} observations to code {code} to balance data')
    _update_dataset_size(dataset)


def _update_dataset_size(dataset):
    """This updates properties of the dataset to keep the length consistent,
    because they are used by the __len__() functions. Not clear what the
    different roles of the two different properties is. See also:
    
    - https://github.com/pytorch/tnt/blob/master/torchnet/dataset/concatdataset.py
    
    This modifies the dataset in-place.

    Parameters
    ----------
    dataset
    """
    dataset.cum_sizes = np.cumsum([len(x) for x in dataset.datasets])
    dataset.cumulative_sizes = dataset.cum_sizes


def _unbalance_dataset(dataset):
    """Removes copied datasets that were add during balancing of the data. This
    modifies the dataset in-place.

    Parameters
    ----------
    dataset
    """
    to_remove = []
    for d in dataset.datasets:
        if hasattr(d, '_is_balance_copy') and d._is_balance_copy:
            to_remove.append(d)
    if not to_remove:
        return
    logger.info(f'removing {len(to_remove)} observations to unbalance data')
    for d in to_remove:
        dataset.datasets.remove(d)
    _update_dataset_size(dataset)


def _build_dataset(epochs, metadata, factors, window_size_samples,
                  window_stride_samples):
    """Creates a dataset that is suitable for braindecode from an Epochs
    object. This indirectly implements 'cropped decoding' as explained on
    the braindecode website and accompanying paper.
    
    Parameters
    ----------
    epochs: Epochs
        The Epochs object to split
    metadata: DataFrame
        The metadata that belongs to epochs
    factors: str or list
        A factor or list of factors by which the data should be split.
    window_size_samples: int
        The length of the window to sample from the Epochs object. This should
        be slightly shorter than the actual Epochs to allow for jittered
        samples to be taken from the purpose of 'cropped decoding'.
    window_stride_samples: int
        The number of samples to jitter around the window for the purpose of
        cropped decoding.
    
    Returns
    -------
    tuple
        A (dataset, labels) tuple, where the labels are as returned by
        split_epochs()
    """
    epochs_list, labels = _split_epochs(epochs, metadata, factors)
    dataset = create_from_mne_epochs(
        epochs_list,
        window_size_samples=window_size_samples,
        window_stride_samples=window_stride_samples,
        drop_last_window=False
    )
    _balance_dataset(dataset)
    return dataset, labels



"""
This file organizes raw EEG & eye-tracking data according to BIDS format.

Assumptions:
-   all EEG files (.eeg, .vhdr, .vmrk) 
    are named in a 'Subject-00X-timestamp' format (e.g. Subject-002-[2022.06.12-14.35.46].eeg)
-   eye-tracking files (.edf)
    are named in a 'sub_X format' (e.g. sub_2.edf)
"""


from pathlib import Path
import argparse
import shutil
import re
import os


def data2bids():
    """
    this function organizes raw eeg&eye-tracking data according to BIDS format
    Arguments:
        source_path: Path where raw data is stored; 
        default: current working directory
        type: str
        
        target_path: Path where BIDS-organized data will be saved
        default: current working directory
        type: str
        
        subjects: List of subject numbers divided by comma (e.g. '1,2,3,4,5')
        type: str
        
        task: Name of the task; default: experimental task
        type: str
    """
    #Parse command-line arguments
    parser = argparse.ArgumentParser()
    parser.add_argument('--source-path', type=str,
                        help='source data path')
    parser.add_argument('--target-path', type=str,
                        help='target data path')
    parser.add_argument('-s','--subjects', type=str,
                        help='subjects id numbers')
    parser.add_argument('-t','--task', type=str,
                        help='task name')
    #Do the parsing
    args = parser.parse_args()
    #If source path was not specified...
    if args.source_path is None:
        #source path would be a current working directory
        src_path = Path('.')
    else:
        #else source path would be the specified one
        src_path = Path(args.source_path)
    #If target path was not specified...
    if args.target_path is None:
        #target path would be a current working directory
        tgt_path = Path('.')
    else:
        #else target path would be the specified one
        tgt_path = Path(args.target_path)
    #If task name was not specified...
    if args.task is None:
         #default name would be assigned
        task = 'experimental_task'
    else:
        #else specified task name would be assigned
        task = args.task
    subjects = [int(sub) for sub in args.subjects.split(',')]
    for sub in subjects:
        sub_path = tgt_path / Path('sub-{:02d}'.format(sub))
        if sub_path.exists():
            shutil.rmtree(sub_path)
        sub_path.mkdir()
        eye_path = sub_path / Path('eyetracking')
        eye_path.mkdir()
        edf_path = src_path / Path(f'sub_{sub}.edf')
        shutil.copyfile(
            edf_path,
            eye_path / Path('sub-{:02d}_{}_physio.edf'.format(sub, task)))
        eeg_path = sub_path / Path('eeg')
        eeg_path.mkdir()
        for ext in ['.eeg', '.vhdr', '.vmrk']:
            src = list(src_path.glob('Subject-{:03d}-*{}'.format(sub, ext)))[0]
            tgt = eeg_path / Path('sub-{:02d}_{}_eeg{}'.format(sub, task, ext))
            shutil.copyfile(src, tgt)
            if ext != '.eeg':
                with open(tgt, 'r') as fd:
                    content = fd.read()
                content = re.sub(
                    'MarkerFile=.*\n',
                    'MarkerFile=sub-{:02d}_{}_eeg.vmrk\n'.format(sub, task),
                    content)
                content = re.sub(
                    'DataFile=.*\n',
                    'DataFile=sub-{:02d}_{}_eeg.eeg\n'.format(sub, task),
                    content)
                with open(tgt, 'w') as fd:
                    fd.write(content)



import numpy as np

TRIAL_TRIGGERS = list(range(128, 256))
TRIGGER_ANNOTATION_PREFIX = 'Stimulus/OVTK_StimulationId_Label_'
ZERO_TRIGGER_ANNOTATION = 'Stimulus/OVTK_StimulationId_Label_FF'


def epoch_trigger(events, trigger):
    """Selects a single epoch trigger from a tuple with event information.
    Epoch triggers have values between 1 and 127 (inclusive).

    Parameters
    ----------
    events: tuple
        Event information as returned by `read_subject()`.
    trigger: int
        A trigger code, which is a positive value.

    Returns
    -------
    array:
        A numpy array with events as expected by mne.Epochs().
    """
    events = _validate_events(events)
    if not 128 > trigger > 0:
        raise ValueError('trigger should be a number between 1 and 127')
    triggers = events
    return triggers[triggers[:, 2] == trigger]
    

def trial_trigger(events):
    """Selects all trial triggers from event information. Trial triggers have
    values between 128 and 255 (inclusive).
    
    Parameters
    ----------
    events: tuple
        Event information as returned by `read_subject()`.

    Returns
    -------
    array:
        A numpy array with events as expected by mne.Epochs().
    """    
    events = _validate_events(events)
    triggers = events
    return triggers[triggers[:, 2] >= 128]


def _parse_triggers(label):
    """An internal function that converts labels as stored in our data."""
    if not label.startswith(TRIGGER_ANNOTATION_PREFIX):
        return None
    if label == ZERO_TRIGGER_ANNOTATION:
        return None
    return 255 - int(label[-2:], 16)
    
    
def _validate_events(events):
    """Checks whether the events are in the correct format."""
    if isinstance(events, tuple):
        if len(events) != 2:
            raise ValueError(
                'events should be a numpy array or tuple of length 2')
        events = events[0]
    if not isinstance(events, np.ndarray):
        raise TypeError('events should be a numpy array')
    codes = events[:, 2]
    if np.any((codes < 1) | (codes > 255)):
        raise ValueError('trigger codes should be values between 1 and 255')
    return events



"""A Python module for reading concurrently recorded EEG and eye-tracking data,
and parsing this data into convenient objects for further analysis.
"""
import logging
from ._triggers import trial_trigger, epoch_trigger
from ._parsing import read_subject
from ._custom_epochs import PupilEpochs, autoreject_epochs, epochs_to_series, \
    tfr_to_surface


__version__ = '0.12.2'
logger = logging.getLogger('eeg_eyetracking_parser')
logger.info(f'eeg_eyetracking_parser {__version__}')
